import tempfile
import unittest
from pathlib import Path
import numpy as np
import sys

from datalad.api import install, Dataset
from parameterized import parameterized

import spikeextractors as se
from spikeextractors.testing import check_recordings_equal, check_sortings_equal

run_local = False
test_nwb = True
test_caching = True

if sys.platform == "linux" or run_local:
    class TestNwbConversions(unittest.TestCase):

        def setUp(self):
            pt = Path.cwd() / 'ephy_testing_data'
            if pt.exists():
                self.dataset = Dataset(pt)
            else:
                self.dataset = install('https://gin.g-node.org/NeuralEnsemble/ephy_testing_data')
            self.savedir = Path(tempfile.mkdtemp())

        @parameterized.expand([
            (
                se.AxonaRecordingExtractor,
                "axona",
                dict(filename=str(Path.cwd() / "ephy_testing_data" / "axona" / "axona_raw.set"))
            ),
            (
                se.BlackrockRecordingExtractor,
                "blackrock/blackrock_2_1",
                dict(
                    filename=str(Path.cwd() / "ephy_testing_data" / "blackrock" / "blackrock_2_1" / "l101210-001"),
                    seg_index=0,
                    nsx_to_load=5
                )
            ),
            (
                se.IntanRecordingExtractor,
                "intan",
                dict(file_path=Path.cwd() / "ephy_testing_data" / "intan" / "intan_rhd_test_1.rhd")
            ),
            (
                se.IntanRecordingExtractor,
                "intan",
                dict(file_path=Path.cwd() / "ephy_testing_data" / "intan" / "intan_rhs_test_1.rhs")
            ),
            # Klusta - no .prm config file in ephy_testing
            # (
            #     se.KlustaRecordingExtractor,
            #     "kwik",
            #     dict(folder_path=Path.cwd() / "ephy_testing_data" / "kwik")
            # ),
            (
                se.MEArecRecordingExtractor,
                "mearec/mearec_test_10s.h5",
                dict(file_path=Path.cwd() / "ephy_testing_data" / "mearec" / "mearec_test_10s.h5")
            ),
            (
                se.NeuralynxRecordingExtractor,
                "neuralynx/Cheetah_v5.7.4/original_data",
                dict(
                    dirname=Path.cwd() / "ephy_testing_data" / "neuralynx" / "Cheetah_v5.7.4" / "original_data",
                    seg_index=0
                )
            ),
            (
                se.NeuroscopeRecordingExtractor,
                "neuroscope/test1",
                dict(file_path=Path.cwd() / "ephy_testing_data" / "neuroscope" / "test1" / "test1.dat")
            ),
            # Nixio - RuntimeError: Cannot open non-existent file in ReadOnly mode!
            # (
            #     se.NIXIORecordingExtractor,
            #     "nix",
            #     dict(file_path=str(Path.cwd() / "ephy_testing_data" / "neoraw.nix"))
            # ),
            (
                se.OpenEphysRecordingExtractor,
                "openephys/OpenEphys_SampleData_1",
                dict(folder_path=Path.cwd() / "ephy_testing_data" / "openephys" / "OpenEphys_SampleData_1")
            ),
            (
                se.OpenEphysRecordingExtractor,
                "openephysbinary/v0.4.4.1_with_video_tracking",
                dict(folder_path=Path.cwd() / "ephy_testing_data" / "openephysbinary" / "v0.4.4.1_with_video_tracking")
            ),
            (
                se.OpenEphysNPIXRecordingExtractor,
                "openephysbinary/v0.5.3_two_neuropixels_stream",
                dict(
                    folder_path=Path.cwd() / "ephy_testing_data" / "openephysbinary" / "v0.5.3_two_neuropixels_stream"
                                / "Record_Node_107")
            ),
            (
                se.NeuropixelsDatRecordingExtractor,
                "openephysbinary/v0.5.3_two_neuropixels_stream",
                dict(
                    file_path=Path.cwd() / "ephy_testing_data" / "openephysbinary" / "v0.5.3_two_neuropixels_stream" /
                              "Record_Node_107" / "experiment1" / "recording1" / "continuous" /
                              "Neuropix-PXI-116.0" / "continuous.dat",
                    settings_file=Path.cwd() / "ephy_testing_data" / "openephysbinary" /
                                  "v0.5.3_two_neuropixels_stream" / "Record_Node_107" / "settings.xml")
            ),
            (
                se.PhyRecordingExtractor,
                "phy/phy_example_0",
                dict(folder_path=Path.cwd() / "ephy_testing_data" / "phy" / "phy_example_0")
            ),
            # Plexon - AssertionError: This file have several channel groups spikeextractors support only one groups
            # (
            #     se.PlexonRecordingExtractor,
            #     "plexon",
            #     dict(filename=Path.cwd() / "ephy_testing_data" / "plexon" / "File_plexon_2.plx")
            # ),
            (
                    se.CEDRecordingExtractor,
                    "spike2/m365_1sec.smrx",
                    dict(
                        file_path=Path.cwd() / "ephy_testing_data" / "spike2" / "m365_1sec.smrx",
                        smrx_channel_ids=range(10)
                    )
            ),
            (
                se.SpikeGLXRecordingExtractor,
                "spikeglx/Noise4Sam_g0",
                dict(
                    file_path=Path.cwd() / "ephy_testing_data" / "spikeglx" / "Noise4Sam_g0" / "Noise4Sam_g0_imec0" /
                    "Noise4Sam_g0_t0.imec0.ap.bin"
                )
            )
        ])
        def test_convert_recording_extractor_to_nwb(self, se_class, dataset_path, se_kwargs):
            print(f"\n\n\n TESTING {se_class.extractor_name}...")
            dataset_stem = Path(dataset_path).stem
            self.dataset.get(dataset_path)
            recording = se_class(**se_kwargs)

            # # test writing to NWB
            if test_nwb:
                nwb_save_path = self.savedir / f"{se_class.__name__}_test_{dataset_stem}.nwb"
                se.NwbRecordingExtractor.write_recording(recording, nwb_save_path, write_scaled=True)
                nwb_recording = se.NwbRecordingExtractor(nwb_save_path)
                check_recordings_equal(recording, nwb_recording, check_times=False)

                if recording.has_unscaled:
                    nwb_save_path_unscaled = self.savedir / f"{se_class.__name__}_test_{dataset_stem}_unscaled.nwb"
                    if np.all(recording.get_channel_offsets() == 0):
                        se.NwbRecordingExtractor.write_recording(recording, nwb_save_path_unscaled, write_scaled=False)
                        nwb_recording = se.NwbRecordingExtractor(nwb_save_path_unscaled)
                        check_recordings_equal(recording, nwb_recording, return_scaled=False, check_times=False)
                        # Skip check when NWB converts uint to int
                        if recording.get_dtype(return_scaled=False) == nwb_recording.get_dtype(return_scaled=False):
                            check_recordings_equal(recording, nwb_recording, return_scaled=True, check_times=False)

            # test caching
            if test_caching:
                rec_cache = se.CacheRecordingExtractor(recording)
                check_recordings_equal(recording, rec_cache)
                if recording.has_unscaled:
                    rec_cache_unscaled = se.CacheRecordingExtractor(recording, return_scaled=False)
                    check_recordings_equal(recording, rec_cache_unscaled, return_scaled=False)
                    check_recordings_equal(recording, rec_cache_unscaled, return_scaled=True)

        @parameterized.expand([
            (
                se.BlackrockSortingExtractor,
                "blackrock/blackrock_2_1",
                dict(
                    filename=str(Path.cwd() / "ephy_testing_data" / "blackrock" / "blackrock_2_1" / "l101210-001"),
                    seg_index=0,
                    nsx_to_load=5
                 )
            ),
            (
                se.KlustaSortingExtractor,
                "kwik",
                dict(file_or_folder_path=Path.cwd() / "ephy_testing_data" / "kwik" / "neo.kwik")
            ),
            # Neuralynx - units_ids = nwbfile.units.id[:] - AttributeError: 'NoneType' object has no attribute 'id'
            # Is the GIN data OK? Or are there no units?
            # (
            #     se.NeuralynxSortingExtractor,
            #     "neuralynx/Cheetah_v5.7.4/original_data",
            #     dict(
            #         dirname=Path.cwd() / "ephy_testing_data" / "neuralynx" / "Cheetah_v5.7.4" / "original_data",
            #         seg_index=0
            #     )
            # ),
            # NIXIO - return [int(da.label) for da in self._spike_das]
            # TypeError: int() argument must be a string, a bytes-like object or a number, not 'NoneType'
            # (
            #     se.NIXIOSortingExtractor,
            #     "nix/nixio_fr.nix",
            #     dict(file_path=str(Path.cwd() / "ephy_testing_data" / "nix" / "nixio_fr.nix"))
            # ),
            (
                se.MEArecSortingExtractor,
                "mearec/mearec_test_10s.h5",
                dict(file_path=Path.cwd() / "ephy_testing_data" / "mearec" / "mearec_test_10s.h5")
            ),
            (
                se.PhySortingExtractor,
                "phy/phy_example_0",
                dict(folder_path=Path.cwd() / "ephy_testing_data" / "phy" / "phy_example_0")
            ),
            (
                se.PlexonSortingExtractor,
                "plexon",
                dict(filename=Path.cwd() / "ephy_testing_data" / "plexon" / "File_plexon_2.plx")
            ),
            (
                se.SpykingCircusSortingExtractor,
                "spykingcircus/spykingcircus_example0",
                dict(
                    file_or_folder_path=Path.cwd() / "ephy_testing_data" / "spykingcircus" / "spykingcircus_example0" /
                                        "recording"
                )
            ),
            # # Tridesclous - dataio error, GIN data is not correct?
            # (
            #     se.TridesclousSortingExtractor,
            #     "tridesclous/tdc_example0",
            #     dict(folder_path=Path.cwd() / "ephy_testing_data" / "tridesclous" / "tdc_example0")
            # )
        ])
        def test_convert_sorting_extractor_to_nwb(self, se_class, dataset_path, se_kwargs):
            print(f"\n\n\n TESTING {se_class.extractor_name}...")
            dataset_stem = Path(dataset_path).stem
            self.dataset.get(dataset_path)

            sorting = se_class(**se_kwargs)
            sf = sorting.get_sampling_frequency()
            if sf is None:  # need to set dummy sampling frequency since no associated acquisition in file
                sf = 30000
                sorting.set_sampling_frequency(sf)

            if test_nwb:
                nwb_save_path = self.savedir / f"{se_class.__name__}_test_{dataset_stem}.nwb"
                se.NwbSortingExtractor.write_sorting(sorting, nwb_save_path)
                nwb_sorting = se.NwbSortingExtractor(nwb_save_path, sampling_frequency=sf)
                check_sortings_equal(sorting, nwb_sorting)

            if test_caching:
                sort_cache = se.CacheSortingExtractor(sorting)
                check_sortings_equal(sorting, sort_cache)


if __name__ == '__main__':
    unittest.main()



from .neobaseextractor import NeoBaseRecordingExtractor, NeoBaseSortingExtractor
from pathlib import Path
from typing import Union, Optional

try:
    import neo
    HAVE_NEO = True
except ImportError:
    HAVE_NEO = False


PathType = Union[str, Path]


class BlackrockRecordingExtractor(NeoBaseRecordingExtractor):
    """
    The Blackrock extractor is wrapped from neo.rawio.BlackrockRawIO.
    
    Parameters
    ----------
    filename: str
        The Blackrock file (.ns1, .ns2, .ns3, .ns4m .ns4, or .ns6)
    block_index: None or int
        If the underlying dataset have several blocks the index must be specified.
    seg_index_index: None or int
        If the underlying dataset have several segments the index must be specified.
    
    """
    extractor_name = 'BlackrockRecording'
    mode = 'file'
    installed = HAVE_NEO
    NeoRawIOClass = 'BlackrockRawIO'

    def __init__(self, filename: PathType, nsx_to_load: Optional[int] = None, block_index: Optional[int] = None, 
                 seg_index: Optional[int] = None, **kwargs):
        super().__init__(filename=filename, nsx_to_load=nsx_to_load, 
                         block_index=block_index, seg_index=seg_index, **kwargs)


class BlackrockSortingExtractor(NeoBaseSortingExtractor):
    """
    The Blackrock extractor is wrapped from neo.rawio.BlackrockRawIO.

    Parameters
    ----------
    filename: str
        The Blackrock file (.nev)
    block_index: None or int
        If the underlying dataset have several blocks the index must be specified.
    seg_index_index: None or int
        If the underlying dataset have several segments the index must be specified.

    """
    extractor_name = 'BlackrockSorting'
    mode = 'file'
    installed = HAVE_NEO
    NeoRawIOClass = 'BlackrockRawIO'

    def __init__(self, filename: PathType, nsx_to_load: Optional[int] = None,
                 block_index: Optional[int] = None, seg_index: Optional[int] = None, **kwargs):
        super().__init__(filename=filename, nsx_to_load=nsx_to_load,
                         block_index=block_index, seg_index=seg_index, **kwargs)



from spikeextractors import RecordingExtractor, SortingExtractor
from pathlib import Path
import numpy as np
from spikeextractors.extraction_tools import check_get_traces_args, check_get_unit_spike_train, check_get_ttl_args
from distutils.version import StrictVersion
import warnings


try:
    import pyopenephys
    HAVE_OE = True

    if pyopenephys.__version__ >= StrictVersion("1.1.2"):
        HAVE_OE_11 = True
    else:
        warnings.warn("pyopenephys>=1.1.2 should be installed. Support for older versions will be removed in "
                      "future releases. Install with:\n\n pip install --upgrade pyopenephys\n\n")
        HAVE_OE_11 = False
except ImportError:
    HAVE_OE = False
    HAVE_OE_11 = False

extractors_dir = Path(__file__).parent.parent


class OpenEphysRecordingExtractor(RecordingExtractor):
    extractor_name = 'OpenEphysRecording'
    has_default_locations = False
    has_unscaled = True
    installed = HAVE_OE  # check at class level if installed or not
    is_writable = False
    mode = 'folder'
    installation_mesg = "To use the OpenEphys extractor, install pyopenephys: \n\n pip install pyopenephys\n\n"

    def __init__(self, folder_path, experiment_id=0, recording_id=0):
        assert self.installed, self.installation_mesg
        RecordingExtractor.__init__(self)
        self._recording_file = folder_path

        self._fileobj = pyopenephys.File(folder_path)
        self._recording = self._fileobj.experiments[experiment_id].recordings[recording_id]
        self._set_analogsignal(self._recording.analog_signals[0])
        self._kwargs = {'folder_path': str(Path(folder_path).absolute()), 'experiment_id': experiment_id,
                        'recording_id': recording_id}

    def _set_analogsignal(self, analogsignals):
        self._analogsignals = analogsignals
        # Set gains: int16 to uV
        if HAVE_OE_11:
            self.set_channel_gains(gains=self._analogsignals.gains)
        else:
            self.set_channel_gains(gains=self._analogsignals.gain)

    def get_channel_ids(self):
        if HAVE_OE_11:
            return list(self._analogsignals.channel_ids)
        else:
            return list(range(self._analogsignals.signal.shape[0]))

    def get_num_frames(self):
        return self._analogsignals.signal.shape[1]

    def get_sampling_frequency(self):
        if HAVE_OE_11:
            return self._analogsignals.sample_rate
        else:
            return float(self._recording.sample_rate.rescale('Hz').magnitude)

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        return self._analogsignals.signal[channel_ids, start_frame:end_frame]

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        channels = [np.unique(ev.channels)[0] for ev in self._recording.events]
        assert channel_id in channels, f"Specified 'channel' not found. Available channels are {channels}"
        ev = self._recording.events[channels.index(channel_id)]

        ttl_frames = (ev.times.rescale("s") * self.get_sampling_frequency()).magnitude.astype(int)
        ttl_states = np.sign(ev.channel_states)
        ttl_valid_idxs = np.where((ttl_frames >= start_frame) & (ttl_frames < end_frame))[0]
        return ttl_frames[ttl_valid_idxs], ttl_states[ttl_valid_idxs]


class OpenEphysNPIXRecordingExtractor(OpenEphysRecordingExtractor):
    extractor_name = 'OpenEphysNPIXRecording'
    has_default_locations = False
    installed = HAVE_OE_11  # check at class level if installed or not
    is_writable = False
    mode = 'folder'
    installation_mesg = "To use the OpenEphys extractor, " \
                        "install pyopenephys >= 1.1: \n\n pip install pyopenephys>=1.1\n\n"

    def __init__(self, folder_path, experiment_id=0, recording_id=0, stream="AP"):
        assert self.installed, self.installation_mesg
        assert stream.upper() in ["AP", "LFP"]
        OpenEphysRecordingExtractor.__init__(self, folder_path, experiment_id, recording_id)

        analogsignals = self._recording.analog_signals
        for analog in analogsignals:
            channel_names = analog.channel_names
            if np.all([stream.upper() in chan for chan in channel_names]):
                self._set_analogsignal(analog)
                # load neuropixels locations
                channel_locations = np.loadtxt(extractors_dir / 'neuropixelsdatrecordingextractor' /
                                               'channel_positions_neuropixels.txt').T
                # get correct channel ID from channel name (e.g. AP32 --> 32)
                channel_ids = [int(chan_name[chan_name.find(stream.upper())+len(stream):]) - 1
                               for chan_name in channel_names]
                locations = channel_locations[channel_ids]
                self.set_channel_locations(locations)
                for i, ch in enumerate(self.get_channel_ids()):
                    self.set_channel_property(ch, "channel_name", channel_names[i])
                break

        self._kwargs = {'folder_path': str(Path(folder_path).absolute()), 'experiment_id': experiment_id,
                        'recording_id': recording_id, 'stream': stream}


class OpenEphysSortingExtractor(SortingExtractor):
    extractor_name = 'OpenEphysSorting'
    installed = HAVE_OE  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the OpenEphys extractor, install pyopenephys: \n\n pip install pyopenephys\n\n"  # error message when not installed

    def __init__(self, folder_path, experiment_id=0, recording_id=0):
        assert self.installed, self.installation_mesg
        SortingExtractor.__init__(self)
        self._recording_file = folder_path
        self._recording = pyopenephys.File(folder_path).experiments[experiment_id].recordings[recording_id]
        self._spiketrains = self._recording.spiketrains
        self._unit_ids = list([np.unique(st.clusters)[0] for st in self._spiketrains])
        self._sampling_frequency = float(self._recording.sample_rate.rescale('Hz').magnitude)

        self._kwargs = {'folder_path': str(Path(folder_path).absolute()), 'experiment_id': experiment_id,
                        'recording_id': recording_id}

    def get_unit_ids(self):
        return self._unit_ids

    @check_get_unit_spike_train
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):

        st = self._spiketrains[unit_id]
        inds = np.where((start_frame <= (st.times * self._recording.sample_rate)) &
                        ((st.times * self._recording.sample_rate) < end_frame))
        return (st.times[inds] * self._recording.sample_rate).magnitude



from .readSGLX import readMeta, SampRate, makeMemMapRaw, GainCorrectIM, GainCorrectNI, ExtractDigital
import numpy as np
from pathlib import Path

from spikeextractors import RecordingExtractor
from spikeextractors.extraction_tools import check_get_traces_args, check_get_ttl_args
import re

class SpikeGLXRecordingExtractor(RecordingExtractor):
    """
    RecordingExtractor from a SpikeGLX Neuropixels file

    Parameters
    ----------
    file_path: str or Path
        Path to the ap.bin, lf.bin, or nidq.bin file
    dtype: str
        'int16' or 'float'. If 'float' is selected, the returned traces are converted to uV
    x_pitch: int
        The x pitch of the probe (default 16)
    y_pitch: int
        The y pitch of the probe (default 20)
    """
    extractor_name = 'SpikeGLXRecording'
    has_default_locations = True
    has_unscaled = True
    installed = True  # check at class level if installed or not
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the SpikeGLXRecordingExtractor run:\n\n pip install mtscomp\n\n"  # error message when not installed

    def __init__(self, file_path: str, x_pitch: int = 32, y_pitch: int = 20):
        RecordingExtractor.__init__(self)
        self._npxfile = Path(file_path)
        self._basepath = self._npxfile.parents[0]
        
        # Gets file type: 'imec0.ap', 'imec0.lf' or 'nidq'
        assert re.search(r'imec[0-9]*.(ap|lf){1}.bin$', self._npxfile.name) or  'nidq' in self._npxfile.name, \
               "'file_path' can be an imec.ap, imec.lf, imec0.ap, imec0.lf, or nidq file"
           
        if 'ap.bin' in str(self._npxfile):
            rec_type = "ap"
            self.is_filtered = True
        elif 'lf.bin' in str(self._npxfile):
            rec_type = "lf"
        else:
            rec_type = "nidq"
        aux = self._npxfile.stem.split('.')[-1]
        if aux == 'nidq':
            self._ftype = aux
        else:
            self._ftype = self._npxfile.stem.split('.')[-2] + '.' + aux

        # Metafile
        self._metafile = self._basepath.joinpath(self._npxfile.stem+'.meta')
        if not self._metafile.exists():
            raise Exception("'meta' file for '"+self._ftype+"' traces should be in the same folder.")
        # Read in metadata, returns a dictionary
        meta = readMeta(self._npxfile)
        self._meta = meta

        # Traces in 16-bit format
        self._raw = makeMemMapRaw(self._npxfile, meta)  # [chanList, firstSamp:lastSamp+1]

        # sampling rate and ap channels
        self._sampling_frequency = SampRate(meta)

        tot_chan, ap_chan, lfp_chan, locations, channel_ids, channel_names \
            = _parse_spikeglx_metafile(self._metafile,
                                       x_pitch=x_pitch,
                                       y_pitch=y_pitch,
                                       rec_type=rec_type)
        if rec_type in ("ap", "lf"):
            self._channels = channel_ids
            # locations
            if len(locations) > 0:
                self.set_channel_locations(locations)
            if len(channel_names) > 0:
                if len(channel_names) == len(self._channels):
                    for i, ch in enumerate(self._channels):
                        self.set_channel_property(ch, "channel_name", channel_names[i])

            if rec_type == "ap":
                if ap_chan < tot_chan:
                    self._timeseries = self._raw[0:ap_chan, :]
            elif rec_type == "lf":
                if lfp_chan < tot_chan:
                    self._timeseries = self._raw[0:lfp_chan, :]
        else:
            # nidq
            self._channels = list(range(int(tot_chan)))
            self._timeseries = self._raw

        # get gains
        if meta['typeThis'] == 'imec':
            gains = GainCorrectIM(self._timeseries, self._channels, meta)
        elif meta['typeThis'] == 'nidq':
            gains = GainCorrectNI(self._timeseries, self._channels, meta)

        # set gains - convert from int16 to uVolt
        self.set_channel_gains(gains=gains*1e6, channel_ids=self._channels)
        self._kwargs = {'file_path': str(Path(file_path).absolute()),
                        'x_pitch': x_pitch, 'y_pitch': y_pitch}

    def get_channel_ids(self):
        return self._channels

    def get_num_frames(self):
        return self._timeseries.shape[1]

    def get_sampling_frequency(self):
        return self._sampling_frequency

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        channel_idxs = np.array([self.get_channel_ids().index(ch) for ch in channel_ids])
        if np.all(channel_ids == self.get_channel_ids()):
            traces = self._timeseries[:, start_frame:end_frame]
        else:
            if np.all(np.diff(channel_idxs) == 1):
                traces = self._timeseries[channel_idxs[0]:channel_idxs[0]+len(channel_idxs), start_frame:end_frame]
            else:
                # This block of the execution will return the data as an array, not a memmap
                traces = self._timeseries[channel_idxs, start_frame:end_frame]

        return traces

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        channel = [channel_id]
        dw = 0
        dig = ExtractDigital(self._raw, firstSamp=start_frame, lastSamp=end_frame, dwReq=dw, dLineList=channel,
                             meta=self._meta)
        dig = np.squeeze(dig)
        diff_dig = np.diff(dig.astype(int))

        rising = np.where(diff_dig > 0)[0] + start_frame
        falling = np.where(diff_dig < 0)[0] + start_frame

        ttl_frames = np.concatenate((rising, falling))
        ttl_states = np.array([1] * len(rising) + [-1] * len(falling))
        sort_idxs = np.argsort(ttl_frames)
        return ttl_frames[sort_idxs], ttl_states[sort_idxs]


def _parse_spikeglx_metafile(metafile, x_pitch, y_pitch, rec_type):
    tot_channels = None
    ap_channels = None
    lfp_channels = None

    y_offset = 20
    x_offset = 11

    locations = []
    channel_names = []
    channel_ids = []
    with Path(metafile).open() as f:
        for line in f.readlines():
            if 'nSavedChans' in line:
                tot_channels = int(line.split('=')[-1])
            if 'snsApLfSy' in line:
                ap_channels = int(line.split('=')[-1].split(',')[0].strip())
                lfp_channels = int(line.split(',')[-2].strip())
            if 'imSampRate' in line:
                fs = float(line.split('=')[-1])
            if rec_type in ("ap", "lf"):
                if 'snsChanMap' in line:
                    map = line.split('=')[-1]
                    chans = map.split(')')[1:]
                    for chan in chans:
                        chan_name = chan[1:].split(';')[0]
                        if rec_type == "ap":
                            if "AP" in chan_name:
                                channel_names.append(chan_name)
                                chan_id = int(chan_name[2:])
                                channel_ids.append(chan_id)
                        elif rec_type == "lf":
                            if "LF" in chan_name:
                                channel_names.append(chan_name)
                                chan_id = int(chan_name[2:])
                                channel_ids.append(chan_id)
                if 'snsShankMap' in line:
                    map = line.split('=')[-1]
                    chans = map.split(')')[1:]
                    for chan in chans:
                        chan = chan[1:]
                        if len(chan) > 0:
                            x_idx = int(chan.split(':')[1])
                            y_idx = int(chan.split(':')[2])
                            stagger = np.mod(y_idx + 0, 2) * x_pitch / 2
                            x_pos = (1 - x_idx) * x_pitch + stagger + x_offset
                            y_pos = y_idx * y_pitch + y_offset
                            locations.append([x_pos, y_pos])
    return tot_channels, ap_channels, lfp_channels, locations, channel_ids, channel_names



import numpy as np
from pathlib import Path
from distutils.version import StrictVersion
from typing import Union, Optional

from spikeextractors import RecordingExtractor
from spikeextractors.extraction_tools import check_get_traces_args, check_get_ttl_args

DtypeType = Union[str, np.dtype]
OptionalArrayType = Optional[Union[np.ndarray, list]]

try:
    import pyintan
    if StrictVersion(pyintan.__version__) >= '0.3.0':
        HAVE_INTAN = True
    else:
        print("pyintan version requires an update (>=0.3.0). Please upgrade with 'pip install --upgrade pyintan'")
        HAVE_INTAN = False
except ImportError:
    HAVE_INTAN = False


class IntanRecordingExtractor(RecordingExtractor):
    """
    Extracts raw neural recordings from the Intan file format.

    The recording extractor always returns channel IDs starting from 0.

    The recording data will always be returned in the shape of (num_channels, num_frames).

    Parameters
    ----------
    file_path : str
        Path to the .dat file to be extracted.
    dtype : dtype
        The data type used in the binary file.
    verbose : bool, optional
        Print output during pyintan file read.
    """

    extractor_name = 'IntanRecording'
    has_default_locations = False
    has_unscaled = True
    is_writable = False
    mode = "file"
    installed = HAVE_INTAN
    installation_mesg = "To use the Intan extractor, install pyintan: \n\n pip install pyintan\n\n"

    def __init__(self, file_path: str, verbose: bool = False):
        assert self.installed, self.installation_mesg
        RecordingExtractor.__init__(self)
        assert Path(file_path).suffix == '.rhs' or Path(file_path).suffix == '.rhd', \
            "Only '.rhd' and '.rhs' files are supported"
        self._recording_file = file_path
        self._recording = pyintan.File(file_path, verbose)
        self._num_frames = len(self._recording.times)
        self._analog_channels = np.array([
            ch for ch in self._recording._anas_chan
            if all([other_ch not in ch['name'] for other_ch in ['ADC', 'VDD', 'AUX']])
        ])
        self._num_channels = len(self._analog_channels)
        self._channel_ids = list(range(self._num_channels))
        self._fs = float(self._recording.sample_rate.rescale('Hz').magnitude)

        for i, ch in enumerate(self._analog_channels):
            self.set_channel_gains(channel_ids=i, gains=ch['gain'])
            self.set_channel_offsets(channel_ids=i, offsets=ch['offset'])

        self._kwargs = dict(file_path=str(Path(file_path).absolute()), verbose=verbose)

    def get_channel_ids(self):
        return self._channel_ids

    def get_num_frames(self):
        return self._num_frames

    def get_sampling_frequency(self):
        return self._fs

    @check_get_traces_args
    def get_traces(
        self,
        channel_ids: OptionalArrayType = None,
        start_frame: Optional[int] = None,
        end_frame: Optional[int] = None,
        return_scaled: bool = True,
    ):
        """
        This function extracts and returns a trace from the recorded data from the
        given channels ids and the given start and end frame. It will return
        traces from within three ranges:

            [start_frame, start_frame+1, ..., end_frame-1]
            [start_frame, start_frame+1, ..., final_recording_frame - 1]
            [0, 1, ..., end_frame-1]
            [0, 1, ..., final_recording_frame - 1]

        if both start_frame and end_frame are given, if only start_frame is
        given, if only end_frame is given, or if neither start_frame or end_frame
        are given, respectively. Traces are returned in a 2D array that
        contains all of the traces from each channel with dimensions
        (num_channels x num_frames). In this implementation, start_frame is inclusive
        and end_frame is exclusive conforming to numpy standards.

        Parameters
        ----------
        start_frame : int, optional
            The starting frame of the trace to be returned (inclusive)
        end_frame : int, optional
            The ending frame of the trace to be returned (exclusive)
        channel_ids : ArrayType, optional
            A list or 1D array of channel ids (ints) from which each trace will be
            extracted
        return_scaled : bool, optional
            If True, traces are returned after scaling (using gain/offset). If False, the raw traces are returned.
            Defaults to True.

        Returns
        ----------
        traces: numpy.ndarray
            A 2D array that contains all of the traces from each channel.
            Dimensions are: (num_channels x num_frames)
        """
        channel_idxs = np.array([self._channel_ids.index(ch) for ch in channel_ids])
        return self._recording._read_analog(
            channels=self._analog_channels[channel_idxs],
            i_start=start_frame,
            i_stop=end_frame,
            dtype="uint16"
        ).T

    @check_get_ttl_args
    def get_ttl_events(self, start_frame=None, end_frame=None, channel_id=0):
        channels = [np.unique(ev.channels)[0] for ev in self._recording.digital_in_events]
        assert channel_id in channels, f"Specified 'channel' not found. Available channels are {channels}"
        ev = self._recording.events[channels.index(channel_id)]

        ttl_frames = (ev.times.rescale("s") * self.get_sampling_frequency()).magnitude.astype(int)
        ttl_states = np.sign(ev.channel_states)
        ttl_valid_idxs = np.where((ttl_frames >= start_frame) & (ttl_frames < end_frame))[0]
        return ttl_frames[ttl_valid_idxs], ttl_states[ttl_valid_idxs]



import numpy as np
import warnings

from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor
from spikeextractors.extraction_tools import check_get_traces_args, check_get_unit_spike_train

try:
    import neo

    HAVE_NEO = True
except ImportError:
    HAVE_NEO = False


class _NeoBaseExtractor:
    NeoRawIOClass = None
    installed = HAVE_NEO
    is_writable = False
    has_default_locations = False
    has_unscaled = True
    installation_mesg = "To use the Neo extractors, install Neo: \n\n pip install neo\n\n"

    def __init__(self, block_index=None, seg_index=None, **kargs):
        """
        if block_index is None then check if only one block
        if seg_index is None then check if only one segment

        """
        assert self.installed, self.installation_mesg
        neoIOclass = eval('neo.rawio.' + self.NeoRawIOClass)

        self.neo_reader = neoIOclass(**kargs)
        self.neo_reader.parse_header()

        if block_index is None:
            # auto select first block
            num_block = self.neo_reader.block_count()
            assert num_block == 1, 'This file is multi block spikeextractors support only one segment, please provide block_index='
            block_index = 0

        if seg_index is None:
            # auto select first segment
            num_seg = self.neo_reader.segment_count(block_index)
            assert num_seg == 1, 'This file is multi segment spikeextractors support only one segment, please provide seg_index='
            seg_index = 0

        self.block_index = block_index
        self.seg_index = seg_index
        self._kwargs = kargs
        self._kwargs.update({'seg_index': seg_index, 'block_index': block_index})


class NeoBaseRecordingExtractor(RecordingExtractor, _NeoBaseExtractor):

    def __init__(self, block_index=None, seg_index=None, **kargs):
        RecordingExtractor.__init__(self)
        _NeoBaseExtractor.__init__(self, block_index=block_index, seg_index=seg_index, **kargs)

        if hasattr(self.neo_reader, 'get_group_signal_channel_indexes'):
            # Neo >= 0.9.0
            channel_indexes_list = self.neo_reader.get_group_signal_channel_indexes()
            num_streams = len(channel_indexes_list)
            assert num_streams == 1, 'This file have several channel groups spikeextractors support only one groups'
            self.after_v10 = False
        elif hasattr(self.neo_reader, 'get_group_channel_indexes'):
            # Neo < 0.9.0
            channel_indexes_list = self.neo_reader.get_group_channel_indexes()
            num_streams = len(channel_indexes_list)
            self.after_v10 = False
        elif hasattr(self.neo_reader, 'signal_streams_count'):
            # Neo >= 0.10.0 (not release yet in march 2021)
            num_streams = self.neo_reader.signal_streams_count()
            self.after_v10 = True
        else:
            raise ValueError('Strange neo version. Please upgrade your neo package: pip install --upgrade neo')

        assert num_streams == 1, 'This file have several signal streams spikeextractors support only one streams' \
                                 'Maybe you can use option to select only one stream'

        # spikeextractor for units to be uV implicitly
        # check that units are V, mV or uV
        units = self.neo_reader.header['signal_channels']['units']
        if not np.all(np.isin(units, ['V', 'mV', 'uV'])):
            warnings.warn('Signal units no Volt compatible, assuming scaling as uV')
        self.additional_gain = np.ones(units.size, dtype='float')
        self.additional_gain[units == 'V'] = 1e6
        self.additional_gain[units == 'mV'] = 1e3
        self.additional_gain[units == 'uV'] = 1.
        self.additional_gain[units == ''] = 1.
        self.additional_gain = self.additional_gain.reshape(1, -1)

        # Add channels properties
        header_channels = self.neo_reader.header['signal_channels'][slice(None)]
        self._neo_chan_ids = self.neo_reader.header['signal_channels']['id']

        # In neo there is not guarantee that channel ids are unique.
        # for instance Blacrock can have several times the same chan_id
        # different sampling rate
        # so check it
        assert np.unique(self._neo_chan_ids).size == self._neo_chan_ids.size, 'In this format channel ids are not ' \
                                                                              'unique! Incompatible with SpikeInterface'

        try:
            channel_ids = [int(ch) for ch in self._neo_chan_ids]
        except Exception as e:
            warnings.warn("Could not parse channel ids to int: using linear channel map")
            channel_ids = list(np.arange(len(self._neo_chan_ids)))
        self._channel_ids = channel_ids

        gains = header_channels['gain'] * self.additional_gain[0]
        self.set_channel_gains(gains=gains, channel_ids=self._channel_ids)

        names = header_channels['name']
        for i, ind in enumerate(self._channel_ids):
            self.set_channel_property(channel_id=ind, property_name='name', value=names[i])

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        # in neo rawio channel can acces by names/ids/indexes
        # there is no garranty that ids/names are unique on some formats
        channel_idxs = [self.get_channel_ids().index(ch) for ch in channel_ids]
        neo_chan_ids = self._neo_chan_ids[channel_idxs]
        if self.after_v10:
            raw_traces = self.neo_reader.get_analogsignal_chunk(block_index=self.block_index, seg_index=self.seg_index,
                                                                i_start=start_frame, i_stop=end_frame,
                                                                channel_indexes=None, channel_names=None,
                                                                stream_index=0, channel_ids=neo_chan_ids)
        else:
            raw_traces = self.neo_reader.get_analogsignal_chunk(block_index=self.block_index, seg_index=self.seg_index,
                                                                i_start=start_frame, i_stop=end_frame,
                                                                channel_indexes=None, channel_names=None,
                                                                channel_ids=neo_chan_ids)
        # neo works with (samples, channels) strides
        # so transpose to spikeextractors wolrd
        return raw_traces.transpose()

    def get_num_frames(self):
        # channel_indexes=None means all channels
        if self.after_v10:
            n = self.neo_reader.get_signal_size(self.block_index, self.seg_index, stream_index=0)
        else:
            n = self.neo_reader.get_signal_size(self.block_index, self.seg_index, channel_indexes=None)
        return n

    def get_sampling_frequency(self):
        # channel_indexes=None means all channels
        if self.after_v10:
            sf = self.neo_reader.get_signal_sampling_rate(stream_index=0)
        else:
            sf = self.neo_reader.get_signal_sampling_rate(channel_indexes=None)
        return sf

    def get_channel_ids(self):
        return self._channel_ids


class NeoBaseSortingExtractor(SortingExtractor, _NeoBaseExtractor):
    def __init__(self, block_index=None, seg_index=None, **kargs):
        SortingExtractor.__init__(self)
        _NeoBaseExtractor.__init__(self, block_index=block_index, seg_index=seg_index, **kargs)

        # the sampling frequency is quite tricky because in neo
        # spike are handle in s or ms
        # internally many format do have have the spike time stamps
        # at the same speed as the signal but at a higher clocks speed.
        # here in spikeinterface we need spike index to be at the same speed
        # that signal it do not make sens to have spikes at 50kHz sample
        # when the sig is 10kHz.
        # neo handle this but not spikeextractors

        self._handle_sampling_frequency()

    def _handle_sampling_frequency(self):
        # bacause neo handle spike in times (s or ms) but spikeextractors in frames related to signals.
        # In neo spikes can have diffrents sampling rate than signals so conversion from
        #  signals frames to times is format dependent

        # here the generic case
        #  all channels are in the same neo group so
        if len(self.neo_reader.header['signal_channels']['sampling_rate']) > 0:
            self._neo_sig_sampling_rate = self.neo_reader.header['signal_channels']['sampling_rate'][0]
            self.set_sampling_frequency(self._neo_sig_sampling_rate)
        else:
            warnings.warn("Sampling frequency not found: setting it to 30 kHz")
            self._sampling_frequency = 30000
            self._neo_sig_sampling_rate = self._sampling_frequency

        if hasattr(self.neo_reader, 'get_group_signal_channel_indexes'):
            # Neo >= 0.9.0
            if len(self.neo_reader.get_group_signal_channel_indexes()) > 0:
                self._neo_sig_time_start = self.neo_reader.get_signal_t_start(self.block_index, self.seg_index,
                                                                              channel_indexes=[0])
            else:
                warnings.warn("Start time not found: setting it to 0 s")
                self._neo_sig_time_start = 0
        elif hasattr(self.neo_reader, 'get_group_channel_indexes'):
            # Neo < 0.9.0
            if len(self.neo_reader.get_group_channel_indexes()) > 0:
                self._neo_sig_time_start = self.neo_reader.get_signal_t_start(self.block_index, self.seg_index,
                                                                              channel_indexes=[0])
            else:
                warnings.warn("Start time not found: setting it to 0 s")
                self._neo_sig_time_start = 0
        elif hasattr(self.neo_reader, 'signal_streams_count'):
            # Neo >= 0.10.0 (not release yet in march 2021)
            num_streams = self.neo_reader.signal_streams_count()
            if num_streams > 0:
                self._neo_sig_time_start = self.neo_reader.get_signal_t_start(self.block_index, self.seg_index,
                                                                              stream_index=0)
            else:
                warnings.warn("Start time not found: setting it to 0 s")
                self._neo_sig_time_start = 0
        else:
            raise ValueError('Strange neo version. Please upgrade your neo package: pip install --upgrade neo')

        # For some IOs when there is no signals at inside the dataset this could not work
        # in that case the extractor class must overwrite this method

    def get_unit_ids(self):
        # should be this but this is strings in neo
        #  unit_ids = self.neo_reader.header['unit_channels']['id']

        # in neo unit_ids are string so here we take unit_index
        if 'unit_channels' in self.neo_reader.header:
            unit_ids = np.arange(self.neo_reader.header['unit_channels'].size, dtype='int64')
        elif 'spike_channels' in self.neo_reader.header:
            unit_ids = np.arange(self.neo_reader.header['spike_channels'].size, dtype='int64')
        else:
            raise ValueError('Strange neo version. Please upgrade your neo package: pip install --upgrade neo')
        return unit_ids

    @check_get_unit_spike_train
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        start_frame, end_frame = self._cast_start_end_frame(start_frame, end_frame)
        # this is a string
        #  neo_unit_id = self.neo_reader.header['unit_channels']['id'][unit_id]

        # this is an int
        unit_index = unit_id

        # in neo can be a sample, or hiher sample rate or even float
        try:
            # version >= 0.9.0
            spike_timestamps = self.neo_reader.get_spike_timestamps(block_index=self.block_index,
                                                                    seg_index=self.seg_index,
                                                                    spike_channel_index=unit_index,
                                                                    t_start=None, t_stop=None)
        except TypeError as e:
            # version < 0.9.0
            spike_timestamps = self.neo_reader.get_spike_timestamps(block_index=self.block_index,
                                                                    seg_index=self.seg_index,
                                                                    unit_index=unit_index, t_start=None, t_stop=None)

        if start_frame is not None:
            spike_timestamps = spike_timestamps[spike_timestamps >= start_frame]

        if end_frame is not None:
            spike_timestamps = spike_timestamps[spike_timestamps <= end_frame]

        # convert to second second
        spike_times = self.neo_reader.rescale_spike_timestamp(spike_timestamps, dtype='float64')

        # convert to sample related to recording signals
        spike_indexes = ((spike_times - self._neo_sig_time_start) * self._neo_sig_sampling_rate).astype('int64')
        return spike_indexes



from spikeextractors import RecordingExtractor, SortingExtractor
from spikeextractors.extractors.numpyextractors import NumpyRecordingExtractor
from spikeextractors.extractors.bindatrecordingextractor import BinDatRecordingExtractor
import numpy as np
from pathlib import Path
from spikeextractors.extraction_tools import check_get_unit_spike_train

try:
    import h5py
    HAVE_SCSX = True
except ImportError:
    HAVE_SCSX = False


class SpykingCircusRecordingExtractor(RecordingExtractor):
    """
    RecordingExtractor for a SpykingCircus output folder

    Parameters
    ----------
    folder_path: str or Path
        Path to the output Spyking Circus folder or result folder
    """
    extractor_name = 'SpykingCircusRecording'
    has_default_locations = False
    has_unscaled = False
    installed = True  # check at class level if installed or not
    is_writable = False
    mode = 'folder'
    installation_mesg = ""  # error message when not installed

    def __init__(self, folder_path):
        RecordingExtractor.__init__(self)
        spykingcircus_folder = Path(folder_path)
        listfiles = spykingcircus_folder.iterdir()

        parent_folder = None
        result_folder = None
        for f in listfiles:
            if f.is_dir():
                if any([f_.suffix == '.hdf5' for f_ in f.iterdir()]):
                    parent_folder = spykingcircus_folder
                    result_folder = f

        if parent_folder is None:
            parent_folder = spykingcircus_folder.parent
            for f in parent_folder.iterdir():
                if f.is_dir():
                    if any([f_.suffix == '.hdf5' for f_ in f.iterdir()]):
                        result_folder = spykingcircus_folder

        assert isinstance(parent_folder, Path) and isinstance(result_folder, Path), "Not a valid spyking circus folder"

        params = None
        params_file = None
        for f in parent_folder.iterdir():
            if f.suffix == '.params':
                params = _load_params(f)
                params_file = f
                break
        assert params is not None, "Could not find the .params file"
        recording_name = params_file.stem

        file_format = params["file_format"].lower()
        if file_format == "numpy":
            recording_file = parent_folder / f"{recording_name}.npy"
            self._recording = NumpyRecordingExtractor(recording_file, params["sampling_frequency"])
        elif file_format == "raw_binary":
            recording_file = parent_folder / f"{recording_name}.dat"
            self._recording = BinDatRecordingExtractor(recording_file, sampling_frequency=params["sampling_frequency"],
                                                       numchan=params["nb_channels"], dtype=params["dtype"],
                                                       time_axis=0)
        else:
            raise Exception(f"'file_format' {params['file_format']} is not supported by the "
                            f"SpykingCircusRecordingExtractor")

        if params["mapping"].is_file():
            self._recording = self.load_probe_file(params["mapping"])

        self.params = params
        self._kwargs = {'folder_path': str(Path(folder_path).absolute())}

    def get_channel_ids(self):
        return self._recording.get_channel_ids()

    def get_num_frames(self):
        return self._recording.get_num_frames()

    def get_sampling_frequency(self):
        return self._recording.get_sampling_frequency()

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        return self._recording.get_traces(channel_ids=channel_ids, start_frame=start_frame, end_frame=end_frame,
                                          return_scaled=return_scaled)


class SpykingCircusSortingExtractor(SortingExtractor):
    """
    SortingExtractor for SpykingCircus output folder or file

    Parameters
    ----------
    file_or_folder_path: str or Path
        Path to the output Spyking Circus folder, the result folder, or a specific hdf5 file in the result folder
    load_templates: bool
        If True, templates are loaded from Spyking Circus output
    """
    extractor_name = 'SpykingCircusSorting'
    installed = HAVE_SCSX  # check at class level if installed or not
    is_writable = True
    mode = 'folder'
    installation_mesg = "To use the SpykingCircusSortingExtractor install h5py: \n\n pip install h5py\n\n"

    def __init__(self, file_or_folder_path, load_templates=False):
        assert self.installed, self.installation_mesg
        SortingExtractor.__init__(self)
        file_or_folder_path = Path(file_or_folder_path)

        if file_or_folder_path.is_dir():
            listfiles = file_or_folder_path.iterdir()
            results = None
            parent_folder = None
            result_folder = None
            for f in listfiles:
                if f.is_dir():
                    if any([f_.suffix == '.hdf5' for f_ in f.iterdir()]):
                        parent_folder = file_or_folder_path
                        result_folder = f

            if parent_folder is None:
                parent_folder = file_or_folder_path.parent
                for f in parent_folder.iterdir():
                    if f.is_dir():
                        if any([f_.suffix == '.hdf5' for f_ in f.iterdir()]):
                            result_folder = file_or_folder_path
            # load files
            for f in result_folder.iterdir():
                if 'result.hdf5' in str(f):
                    results = f
                    result_extension = ''
                    base_name = f.name[:f.name.find("result")-1]
                if 'result-merged.hdf5' in str(f):
                    results = f
                    result_extension = '-merged'
                    base_name = f.name[:f.name.find("result")-1]
                    break
        else:
            assert file_or_folder_path.suffix in ['.h5', '.hdf5']
            result_folder = file_or_folder_path.parent
            parent_folder = result_folder.parent
            results = file_or_folder_path
            result_extension = results.stem[results.stem.find("result") + 6:]
            base_name = file_or_folder_path.name[:file_or_folder_path.name.find("result") - 1]

        assert isinstance(parent_folder, Path) and isinstance(result_folder, Path), "Not a valid spyking circus folder"

        # load params
        params = {}
        for f in parent_folder.iterdir():
            if f.suffix == '.params':
                params = _load_params(f)

        if "sampling_frequency" in params.keys():
            self._sampling_frequency = params["sampling_frequency"]

        if results is None:
            raise Exception(f"{file_or_folder_path} is not a spyking circus folder")

        f_results = h5py.File(results, 'r')
        self._spiketrains = []
        self._unit_ids = []
        for temp in f_results['spiketimes'].keys():
            self._spiketrains.append(np.array(f_results['spiketimes'][temp]).astype('int64'))
            self._unit_ids.append(int(temp.split('_')[-1]))

        if load_templates:
            try:
                import scipy
            except:
                raise ImportError("'scipy' is needed to load templates from Spyking Circus")

            filename = result_folder / f"{base_name}.templates{result_extension}.hdf5"
            with h5py.File(filename, 'r', libver='earliest') as f:
                temp_x = f.get('temp_x')[:].ravel()
                temp_y = f.get('temp_y')[:].ravel()
                temp_data = f.get('temp_data')[:].ravel()
                N_e, N_t, nb_templates = f.get('temp_shape')[:].ravel().astype(np.int32)
            templates = scipy.sparse.csc_matrix((temp_data, (temp_x, temp_y)), shape=(N_e * N_t, nb_templates))
            templates = np.array([templates[:, i].toarray().reshape(N_e, N_t) for i in range(templates.shape[1])])

            templates = templates[:len(templates)//2]
            for u_i, unit in enumerate(self.get_unit_ids()):
                self.set_unit_property(unit, 'template', templates[u_i])

        self._kwargs = {'file_or_folder_path': str(Path(file_or_folder_path).absolute()),
                        'load_templates': load_templates}

    def get_unit_ids(self):
        return list(self._unit_ids)

    @check_get_unit_spike_train
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):

        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]

    @staticmethod
    def write_sorting(sorting, save_path):
        assert HAVE_SCSX, SpykingCircusSortingExtractor.installation_mesg
        save_path = Path(save_path)
        if save_path.is_dir():
            save_path = save_path / 'data.result.hdf5'
        elif save_path.suffix == '.hdf5':
            if not str(save_path).endswith('result.hdf5') or not str(save_path).endswith('result-merged.hdf5'):
                raise AttributeError("'save_path' is either a folder or an hdf5 file "
                                     "ending with 'result.hdf5' or 'result-merged.hdf5")
        else:
            save_path.mkdir()
            save_path = save_path / 'data.result.hdf5'
        F = h5py.File(save_path, 'w')
        spiketimes = F.create_group('spiketimes')

        for id in sorting.get_unit_ids():
            spiketimes.create_dataset('tmp_' + str(id), data=sorting.get_unit_spike_train(id))


def _load_params(params_file):
    params = {}
    with params_file.open('r') as f:
        for r in f.readlines():
            if 'sampling_rate' in r:
                sampling_frequency = r.split('=')[-1]
                if '#' in sampling_frequency:
                    sampling_frequency = sampling_frequency[:sampling_frequency.find('#')]
                sampling_frequency = sampling_frequency.strip(" ").strip("\n")
                sampling_frequency = float(sampling_frequency)
                params["sampling_frequency"] = sampling_frequency
            if 'file_format' in r:
                file_format = r.split('=')[-1]
                if '#' in file_format:
                    file_format = file_format[:file_format.find('#')]
                file_format = file_format.strip(" ").strip("\n")
                params["file_format"] = file_format
            if 'nb_channels' in r:
                nb_channels = r.split('=')[-1]
                if '#' in nb_channels:
                    nb_channels = nb_channels[:nb_channels.find('#')]
                nb_channels = nb_channels.strip(" ").strip("\n")
                params["nb_channels"] = int(nb_channels)
            if 'data_dtype' in r:
                dtype = r.split('=')[-1]
                if '#' in dtype:
                    dtype = dtype[:dtype.find('#')]
                dtype = dtype.strip(" ").strip("\n")
                params["dtype"] = dtype
            if 'mapping' in r:
                mapping = r.split('=')[-1]
                if '#' in mapping:
                    mapping = mapping[:mapping.find('#')]
                mapping = mapping.strip(" ").strip("\n")
                params["mapping"] = Path(mapping)
    return params



import numpy as np
from pathlib import Path
import csv
from typing import Union, Optional

from spikeextractors import SortingExtractor, RecordingExtractor
from spikeextractors.extractors.bindatrecordingextractor import BinDatRecordingExtractor
from spikeextractors.extraction_tools import read_python, check_get_unit_spike_train

PathType = Union[str, Path]


class PhyRecordingExtractor(BinDatRecordingExtractor):
    """
    RecordingExtractor for a Phy output folder

    Parameters
    ----------
    folder_path: str or Path
        Path to the output Phy folder (containing the params.py)
    """
    extractor_name = 'PhyRecording'
    has_default_locations = True
    has_unscaled = False
    installed = True  # check at class level if installed or not
    is_writable = False
    mode = 'folder'
    installation_mesg = ""  # error message when not installed

    def __init__(self, folder_path: PathType):
        RecordingExtractor.__init__(self)
        phy_folder = Path(folder_path)

        self.params = read_python(str(phy_folder / 'params.py'))
        datfile = [x for x in phy_folder.iterdir() if x.suffix == '.dat' or x.suffix == '.bin']

        if (phy_folder / 'channel_map_si.npy').is_file():
            channel_map = list(np.squeeze(np.load(phy_folder / 'channel_map_si.npy')))
            assert max(channel_map) < self.params['n_channels_dat'], "Channel map inconsistent with dat file."
        elif (phy_folder / 'channel_map.npy').is_file():
            channel_map = list(np.squeeze(np.load(phy_folder / 'channel_map.npy')))
            assert max(channel_map) < self.params['n_channels_dat'], "Channel map inconsistent with dat file."
        else:
            channel_map = list(range(self.params['n_channels_dat']))

        BinDatRecordingExtractor.__init__(self, datfile[0], sampling_frequency=float(self.params['sample_rate']),
                                          dtype=self.params['dtype'], numchan=self.params['n_channels_dat'],
                                          recording_channels=list(channel_map))

        if (phy_folder / 'channel_groups.npy').is_file():
            channel_groups = np.load(phy_folder / 'channel_groups.npy')
            assert len(channel_groups) == self.get_num_channels()
            self.set_channel_groups(channel_groups)

        if (phy_folder / 'channel_positions.npy').is_file():
            channel_locations = np.load(phy_folder / 'channel_positions.npy')
            assert len(channel_locations) == self.get_num_channels()
            self.set_channel_locations(channel_locations)

        self._kwargs = {'folder_path': str(Path(folder_path).absolute())}


class PhySortingExtractor(SortingExtractor):
    """
    SortingExtractor for a Phy output folder

    Parameters
    ----------
    folder_path: str or Path
        Path to the output Phy folder (containing the params.py)
    exclude_cluster_groups: list (optional)
        List of cluster groups to exclude (e.g. ["noise", "mua"]
    """
    extractor_name = 'PhySorting'
    installed = True  # check at class level if installed or not
    is_writable = False
    mode = 'folder'
    installation_mesg = ""  # error message when not installed

    def __init__(self, folder_path: PathType, exclude_cluster_groups: Optional[list] = None):
        SortingExtractor.__init__(self)
        phy_folder = Path(folder_path)

        spike_times = np.load(phy_folder / 'spike_times.npy')
        spike_templates = np.load(phy_folder / 'spike_templates.npy')

        if (phy_folder / 'spike_clusters.npy').is_file():
            spike_clusters = np.load(phy_folder / 'spike_clusters.npy')
        else:
            spike_clusters = spike_templates

        if (phy_folder / 'amplitudes.npy').is_file():
            amplitudes = np.squeeze(np.load(phy_folder / 'amplitudes.npy'))
        else:
            amplitudes = np.ones(len(spike_times))

        if (phy_folder / 'pc_features.npy').is_file():
            pc_features = np.squeeze(np.load(phy_folder / 'pc_features.npy'))
        else:
            pc_features = None

        clust_id = np.unique(spike_clusters)
        self._unit_ids = list(clust_id)
        spike_times.astype(int)
        self.params = read_python(str(phy_folder / 'params.py'))
        self._sampling_frequency = self.params['sample_rate']

        # set unit quality properties
        csv_tsv_files = [x for x in phy_folder.iterdir() if x.suffix == '.csv' or x.suffix == '.tsv']
        for f in csv_tsv_files:
            if f.suffix == '.csv':
                with f.open() as csv_file:
                    csv_reader = csv.reader(csv_file, delimiter=',')
                    line_count = 0
                    for row in csv_reader:
                        if line_count == 0:
                            tokens = row[0].split("\t")
                            property_name = tokens[1]
                        else:
                            tokens = row[0].split("\t")
                            if int(tokens[0]) in self.get_unit_ids():
                                if 'cluster_group' in str(f):
                                    self.set_unit_property(int(tokens[0]), 'quality', tokens[1])
                                elif property_name == 'chan_grp' or property_name == 'ch_group':
                                    self.set_unit_property(int(tokens[0]), 'group', int(tokens[1]))
                                else:
                                    if isinstance(tokens[1], (int, np.int, float, str)):
                                        self.set_unit_property(int(tokens[0]), property_name, tokens[1])
                            line_count += 1
            elif f.suffix == '.tsv':
                with f.open() as csv_file:
                    csv_reader = csv.reader(csv_file, delimiter='\t')
                    line_count = 0
                    for row in csv_reader:
                        if line_count == 0:
                            property_name = row[1]
                        else:
                            if len(row) == 2:
                                if int(row[0]) in self.get_unit_ids():
                                    if 'cluster_group' in str(f):
                                        self.set_unit_property(int(row[0]), 'quality', row[1])
                                    elif property_name == 'chan_grp' or property_name == 'ch_group':
                                        self.set_unit_property(int(row[0]), 'group', int(row[1]))
                                    else:
                                        if isinstance(row[1], (int, np.int, float, str)) and len(row) == 2:
                                            self.set_unit_property(int(row[0]), property_name, row[1])
                        line_count += 1

        for unit in self.get_unit_ids():
            if 'quality' not in self.get_unit_property_names(unit):
                self.set_unit_property(unit, 'quality', 'unsorted')

        if exclude_cluster_groups is not None:
            if len(exclude_cluster_groups) > 0:
                included_units = []
                for u in self.get_unit_ids():
                    if self.get_unit_property(u, 'quality') not in exclude_cluster_groups:
                        included_units.append(u)
            else:
                included_units = self._unit_ids
        else:
            included_units = self._unit_ids

        original_units = self._unit_ids
        self._unit_ids = included_units
        # set features
        self._spiketrains = []
        for clust in self._unit_ids:
            idx = np.where(spike_clusters == clust)[0]
            self._spiketrains.append(spike_times[idx])
            self.set_unit_spike_features(clust, 'amplitudes', amplitudes[idx])
            if pc_features is not None:
                self.set_unit_spike_features(clust, 'pc_features', pc_features[idx])

        self._kwargs = {'folder_path': str(Path(folder_path).absolute()),
                        'exclude_cluster_groups': exclude_cluster_groups}

    def get_unit_ids(self):
        return list(self._unit_ids)

    @check_get_unit_spike_train
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):
        times = self._spiketrains[self.get_unit_ids().index(unit_id)]
        inds = np.where((start_frame <= times) & (times < end_frame))
        return times[inds]



from .neobaseextractor import NeoBaseRecordingExtractor, NeoBaseSortingExtractor

try:
    import neo
    HAVE_NEO = True
except ImportError:
    HAVE_NEO = False

class SpikeGadgetsRecordingExtractor(NeoBaseRecordingExtractor):
    """
    The spikegadgets extractor is wrapped from neo.rawio.SpikegadgetsRawIO.
    
    Parameters
    ----------
    filename: str
        The spike gadgets file ('rec')
    selected_streams: str
        The id of the stream to load 'trodes' is ephy channels.
        Can also be ECU, ...
    block_index: None or int
        If the underlying dataset have several blocks the index must be specified.
    seg_index_index: None or int
        If the underlying dataset have several segments the index must be specified.
    """    
    extractor_name = 'SpikeGadgetsRecording'
    mode = 'file'
    installed = HAVE_NEO
    NeoRawIOClass = 'SpikeGadgetsRawIO'
    def __init__(self, filename, selected_streams='trodes',**kwargs):
        super().__init__(filename=filename, selected_streams=selected_streams, **kwargs)



from spikeextractors import RecordingExtractor
from spikeextractors import SortingExtractor
from spikeextractors.extraction_tools import write_to_binary_dat_format, check_get_traces_args, \
    check_get_unit_spike_train

import json
import numpy as np
from pathlib import Path
from .mdaio import DiskReadMda, readmda, writemda64, MdaHeader
import shutil


class MdaRecordingExtractor(RecordingExtractor):
    extractor_name = 'MdaRecording'
    has_default_locations = True
    has_unscaled = False
    installed = True  # check at class level if installed or not
    is_writable = True
    mode = 'folder'
    installation_mesg = ""  # error message when not installed

    def __init__(self, folder_path, raw_fname='raw.mda', params_fname='params.json', geom_fname='geom.csv'):
        dataset_directory = Path(folder_path)
        self._dataset_directory = dataset_directory
        timeseries0 = dataset_directory / raw_fname
        self._dataset_params = read_dataset_params(dataset_directory, params_fname)
        self._sampling_frequency = self._dataset_params['samplerate'] * 1.0
        self._timeseries_path = str(timeseries0.absolute())
        geom0 = dataset_directory / geom_fname
        self._geom_fname = geom0
        self._geom = np.loadtxt(self._geom_fname, delimiter=',', ndmin=2)
        X = DiskReadMda(self._timeseries_path)
        if self._geom.shape[0] != X.N1():
            raise Exception(
                'Incompatible dimensions between geom.csv and timeseries file {} <> {}'.format(self._geom.shape[0],
                                                                                               X.N1()))
        self._num_channels = X.N1()
        self._num_timepoints = X.N2()
        RecordingExtractor.__init__(self)
        self.set_channel_locations(self._geom)
        self._kwargs = {'folder_path': str(Path(folder_path).absolute())}

    def get_channel_ids(self):
        return list(range(self._num_channels))

    def get_num_frames(self):
        return self._num_timepoints

    def get_sampling_frequency(self):
        return self._sampling_frequency

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        X = DiskReadMda(self._timeseries_path)
        recordings = X.readChunk(i1=0, i2=start_frame, N1=X.N1(), N2=end_frame - start_frame)
        recordings = recordings[channel_ids, :]
        return recordings

    def write_to_binary_dat_format(self, save_path, time_axis=0, dtype=None, chunk_size=None, chunk_mb=500,
                                   n_jobs=1, joblib_backend='loky', verbose=False):
        """Saves the traces of this recording extractor into binary .dat format.

        Parameters
        ----------
        save_path: str
            The path to the file.
        time_axis: 0 (default) or 1
            If 0 then traces are transposed to ensure (nb_sample, nb_channel) in the file.
            If 1, the traces shape (nb_channel, nb_sample) is kept in the file.
        dtype: dtype
            Type of the saved data. Default float32
        chunk_size: None or int
            Size of each chunk in number of frames.
            If None (default) and 'chunk_mb' is given, the file is saved in chunks of 'chunk_mb' Mb (default 500Mb)
        chunk_mb: None or int
            Chunk size in Mb (default 500Mb)
        n_jobs: int
            Number of jobs to use (Default 1)
        joblib_backend: str
            Joblib backend for parallel processing ('loky', 'threading', 'multiprocessing')
        verbose: bool
            If True, output is verbose
        """
        X = DiskReadMda(self._timeseries_path)
        header_size = X._header.header_size
        if dtype is None or dtype == self.get_dtype():
            try:
                with open(self._timeseries_path, 'rb') as src, open(save_path, 'wb') as dst:
                    src.seek(header_size)
                    shutil.copyfileobj(src, dst)
            except Exception as e:
                print('Error occurred while copying:', e)
                print('Writing to binary')
                write_to_binary_dat_format(self, save_path=save_path, time_axis=time_axis, dtype=dtype,
                                           chunk_size=chunk_size, chunk_mb=chunk_mb, n_jobs=n_jobs,
                                           joblib_backend=joblib_backend, verbose=verbose)
        else:
            write_to_binary_dat_format(self, save_path=save_path, time_axis=time_axis, dtype=dtype,
                                       chunk_size=chunk_size, chunk_mb=chunk_mb, n_jobs=n_jobs,
                                       joblib_backend=joblib_backend, verbose=verbose)

    @staticmethod
    def write_recording(recording, save_path, params=dict(), raw_fname='raw.mda', params_fname='params.json',
                        geom_fname='geom.csv', dtype=None, chunk_size=None, n_jobs=None, chunk_mb=500, verbose=False):
        """
        Writes recording to file in MDA format.

        Parameters
        ----------
        recording: RecordingExtractor
            The recording extractor to be saved
        save_path: str or Path
            The folder in which the Mda files are saved
        params: dictionary
            Dictionary with optional parameters to save metadata. Sampling frequency is appended to this dictionary.
        raw_fname: str
            File name of raw file (default raw.mda)
        params_fname: str
            File name of params file (default params.json)
        geom_fname: str
            File name of geom file (default geom.csv)
        dtype: dtype
            dtype to be used. If None dtype is same as recording traces.
        chunk_size: None or int
            Size of each chunk in number of frames.
            If None (default) and 'chunk_mb' is given, the file is saved in chunks of 'chunk_mb' Mb (default 500Mb)
        n_jobs: int
            Number of jobs to use (Default 1)
        chunk_mb: None or int
            Chunk size in Mb (default 500Mb)
        verbose: bool
            If True, output is verbose
        """
        save_path = Path(save_path)
        save_path.mkdir(parents=True, exist_ok=True)
        save_file_path = save_path / raw_fname
        parent_dir = save_path
        channel_ids = recording.get_channel_ids()
        num_chan = recording.get_num_channels()
        num_frames = recording.get_num_frames()

        geom = recording.get_channel_locations()

        if dtype is None:
            dtype = recording.get_dtype()

        if dtype == 'float':
            dtype = 'float32'
        if dtype == 'int':
            dtype = 'int16'

        with save_file_path.open('wb') as f:
            header = MdaHeader(dt0=dtype, dims0=(num_chan, num_frames))
            header.write(f)
            # takes care of the chunking
            write_to_binary_dat_format(recording, file_handle=f, dtype=dtype, n_jobs=n_jobs, chunk_size=chunk_size,
                                       chunk_mb=chunk_mb, verbose=verbose)

        params["samplerate"] = float(recording.get_sampling_frequency())
        with (parent_dir / params_fname).open('w') as f:
            json.dump(params, f)
        np.savetxt(str(parent_dir / geom_fname), geom, delimiter=',')


class MdaSortingExtractor(SortingExtractor):
    extractor_name = 'MdaSorting'
    installed = True  # check at class level if installed or not
    is_writable = True
    mode = 'file'
    installation_mesg = ""  # error message when not installed

    def __init__(self, file_path, sampling_frequency=None):

        SortingExtractor.__init__(self)
        self._firings_path = file_path
        self._firings = readmda(self._firings_path)
        self._max_channels = self._firings[0, :]
        self._spike_times = self._firings[1, :]
        self._labels = self._firings[2, :]
        self._unit_ids = np.unique(self._labels).astype(int)
        self._sampling_frequency = sampling_frequency
        self._kwargs = {'file_path': str(Path(file_path).absolute()), 'sampling_frequency': sampling_frequency}

    def get_unit_ids(self):
        return list(self._unit_ids)

    @check_get_unit_spike_train
    def get_unit_spike_train(self, unit_id, start_frame=None, end_frame=None):

        inds = np.where(
            (self._labels == unit_id) & (start_frame <= self._spike_times) & (self._spike_times < end_frame))
        return np.rint(self._spike_times[inds]).astype(int)

    @staticmethod
    def write_sorting(sorting, save_path, write_primary_channels=False):
        unit_ids = sorting.get_unit_ids()
        times_list = []
        labels_list = []
        primary_channels_list = []
        for unit_id in unit_ids:
            times = sorting.get_unit_spike_train(unit_id=unit_id)
            times_list.append(times)
            labels_list.append(np.ones(times.shape) * unit_id)
            if write_primary_channels:
                if 'max_channel' in sorting.get_unit_property_names(unit_id):
                    primary_channels_list.append([sorting.get_unit_property(unit_id, 'max_channel')] * times.shape[0])
                else:
                    raise ValueError(
                        "Unable to write primary channels because 'max_channel' spike feature not set in unit " + str(
                            unit_id))
            else:
                primary_channels_list.append(np.zeros(times.shape))
        all_times = _concatenate(times_list)
        all_labels = _concatenate(labels_list)
        all_primary_channels = _concatenate(primary_channels_list)
        sort_inds = np.argsort(all_times)
        all_times = all_times[sort_inds]
        all_labels = all_labels[sort_inds]
        all_primary_channels = all_primary_channels[sort_inds]
        L = len(all_times)
        firings = np.zeros((3, L))
        firings[0, :] = all_primary_channels
        firings[1, :] = all_times
        firings[2, :] = all_labels

        writemda64(firings, save_path)


def _concatenate(list):
    if len(list) == 0:
        return np.array([])
    return np.concatenate(list)


def read_dataset_params(dsdir, params_fname):
    fname1 = dsdir / params_fname
    if not fname1.is_file():
        raise Exception('Dataset parameter file does not exist: ' + fname1)
    with open(fname1) as f:
        return json.load(f)



from spikeextractors.extraction_tools import check_get_traces_args
from spikeextractors.extractors.neoextractors.neobaseextractor import (
    _NeoBaseExtractor, NeoBaseRecordingExtractor)
from spikeextractors import RecordingExtractor
from pathlib import Path
import numpy as np
from typing import Union
import warnings

PathType = Union[Path, str]

try:
    import neo
    from neo.rawio.baserawio import _signal_channel_dtype, _signal_stream_dtype
    HAVE_NEO = True
except ImportError:
    HAVE_NEO = False


class AxonaUnitRecordingExtractor(NeoBaseRecordingExtractor, RecordingExtractor, _NeoBaseExtractor):
    """
    Instantiates a RecordingExtractor from an Axona Unit mode file.

    Since the unit mode format only saves waveform cutouts, the get_traces
    function fills in the rest of the recording with Gaussian uncorrelated
    noise

    Parameters
    ----------

    noise_std: float
        Standard deviation of the Gaussian background noise (default 3)
    """
    extractor_name = 'AxonaUnitRecording'
    mode = 'file'
    NeoRawIOClass = 'AxonaRawIO'

    def __init__(self, noise_std: float = 3, block_index=None, seg_index=None, **kargs):
        RecordingExtractor.__init__(self)
        _NeoBaseExtractor.__init__(self, block_index=block_index, seg_index=seg_index, **kargs)

        # Enforce 1 signal stream (there are 0 raw streams), we will create 1 from waveforms
        signal_streams = self.neo_reader._get_signal_streams_header()
        signal_channels = self.neo_reader._get_signal_chan_header()
        self.neo_reader.header['signal_streams'] = np.array(signal_streams,
                                                            dtype=_signal_stream_dtype)
        self.neo_reader.header['signal_channels'] = np.array(signal_channels,
                                                             dtype=_signal_channel_dtype)

        if hasattr(self.neo_reader, 'get_group_signal_channel_indexes'):
            # Neo >= 0.9.0
            channel_indexes_list = self.neo_reader.get_group_signal_channel_indexes()
            num_streams = len(channel_indexes_list)
            assert num_streams <= 1, 'This file have several channel groups spikeextractors support only one groups'
            self.after_v10 = False
        elif hasattr(self.neo_reader, 'get_group_channel_indexes'):
            # Neo < 0.9.0
            channel_indexes_list = self.neo_reader.get_group_channel_indexes()
            num_streams = len(channel_indexes_list)
            self.after_v10 = False
        elif hasattr(self.neo_reader, 'signal_streams_count'):
            # Neo >= 0.10.0 (not release yet in march 2021)
            num_streams = self.neo_reader.signal_streams_count()
            self.after_v10 = True
        else:
            raise ValueError('Strange neo version. Please upgrade your neo package: pip install --upgrade neo')

        assert num_streams <= 1, 'This file have several signal streams spikeextractors support only one streams' \
                                 'Maybe you can use option to select only one stream'

        # spikeextractor for units to be uV implicitly
        # check that units are V, mV or uV
        units = self.neo_reader.header['signal_channels']['units']
        assert np.all(np.isin(units, ['V', 'mV', 'uV'])), 'Signal units no Volt compatible'
        self.additional_gain = np.ones(units.size, dtype='float')
        self.additional_gain[units == 'V'] = 1e6
        self.additional_gain[units == 'mV'] = 1e3
        self.additional_gain[units == 'uV'] = 1.
        self.additional_gain = self.additional_gain.reshape(1, -1)

        # Add channels properties
        header_channels = self.neo_reader.header['signal_channels'][slice(None)]
        self._neo_chan_ids = self.neo_reader.header['signal_channels']['id']

        # In neo there is not guarantee that channel ids are unique.
        # for instance Blacrock can have several times the same chan_id
        # different sampling rate
        # so check it
        assert np.unique(self._neo_chan_ids).size == self._neo_chan_ids.size, 'In this format channel ids are not ' \
                                                                              'unique! Incompatible with SpikeInterface'

        try:
            channel_ids = [int(ch) for ch in self._neo_chan_ids]
        except Exception as e:
            warnings.warn("Could not parse channel ids to int: using linear channel map")
            channel_ids = list(np.arange(len(self._neo_chan_ids)))
        self._channel_ids = channel_ids

        gains = header_channels['gain'] * self.additional_gain[0]
        self.set_channel_gains(gains=gains, channel_ids=self._channel_ids)

        names = header_channels['name']
        for i, ind in enumerate(self._channel_ids):
            self.set_channel_property(channel_id=ind, property_name='name', value=names[i])

        self._noise_std = noise_std

        # Read channel groups by tetrode IDs
        self.set_channel_groups(groups=[
            tetrode_id - 1 for tetrode_id in self.neo_reader.get_active_tetrode() for _ in range(4)])

        header_channels = self.neo_reader.header['signal_channels'][slice(None)]

        names = header_channels['name']
        channel_ids = self.get_channel_ids()
        for i, ind in enumerate(channel_ids):
            self.set_channel_property(channel_id=ind, property_name='name', value=names[i])

        # Set channel gains for int8 .X Unit data
        gains = self.neo_reader._get_channel_gain(bytes_per_sample=1)[0:len(channel_ids)]
        self.set_channel_gains(gains, channel_ids=channel_ids)

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):

        timebase_sr = int(self.neo_reader.file_parameters['unit']['timebase'].split(' ')[0])
        samples_pre = int(self.neo_reader.file_parameters['set']['file_header']['pretrigSamps'])
        samples_post = int(self.neo_reader.file_parameters['set']['file_header']['spikeLockout'])
        sampling_rate = self.get_sampling_frequency()

        tcmap = self._get_tetrode_channel_table(channel_ids)

        traces = self._noise_std * np.random.randn(len(channel_ids), end_frame - start_frame)
        if return_scaled:
            traces = traces.astype(np.float32)
        else:
            traces = traces.astype(np.int8)

        # Loop through tetrodes and include requested channels in traces
        itrc = 0
        for tetrode_id in np.unique(tcmap[:, 0]):

            channels_oi = tcmap[tcmap[:, 0] == tetrode_id, 2]

            waveforms = self.neo_reader._get_spike_raw_waveforms(
                block_index=0, seg_index=0,
                unit_index=tetrode_id - 1,  # Tetrodes IDs are 1-indexed
                t_start=start_frame / sampling_rate,
                t_stop=end_frame / sampling_rate
            )
            waveforms = waveforms[:, channels_oi, :]
            nch = len(channels_oi)

            spike_train = self.neo_reader._get_spike_timestamps(
                block_index=0, seg_index=0,
                unit_index=tetrode_id - 1,
                t_start=start_frame / sampling_rate,
                t_stop=end_frame / sampling_rate
            )

            # Fill waveforms into traces timestamp by timestamp
            for t, wf in zip(spike_train, waveforms):

                t = int(t // (timebase_sr / sampling_rate))  # timestamps are sampled at higher frequency
                t = t - start_frame
                if (t - samples_pre < 0) and (t + samples_post > traces.shape[1]):
                    traces[itrc:itrc + nch, :] = wf[:, samples_pre - t:traces.shape[1] - (t - samples_pre)]
                elif t - samples_pre < 0:
                    traces[itrc:itrc + nch, :t + samples_post] = wf[:, samples_pre - t:]
                elif t + samples_post > traces.shape[1]:
                    traces[itrc:itrc + nch, t - samples_pre:] = wf[:, :traces.shape[1] - (t - samples_pre)]
                else:
                    traces[itrc:itrc + nch, t - samples_pre:t + samples_post] = wf

            itrc += nch

        return traces

    def get_num_frames(self):
        n = int(self.neo_reader.segment_t_stop(block_index=0, seg_index=0) * self.get_sampling_frequency())
        if self.get_sampling_frequency() == 24000:
            n = n // 2
        return n

    def get_sampling_frequency(self):
        return int(self.neo_reader.header['spike_channels'][0][-1])

    def get_channel_ids(self):
        return self._channel_ids

    def _get_tetrode_channel_table(self, channel_ids):
        '''Create auxiliary np.array with the following columns:
        Tetrode ID, Channel ID, Channel ID within tetrode
        This is useful in `get_traces()`

        Parameters
        ----------
        channel_ids : list
            List of channel ids to include in table

        Returns
        -------
        np.array
            Rows = channels,
            columns = TetrodeID, ChannelID, ChannelID within Tetrode
        '''
        active_tetrodes = self.neo_reader.get_active_tetrode()

        tcmap = np.zeros((len(active_tetrodes) * 4, 3), dtype=int)
        row_id = 0
        for tetrode_id in [int(s[0].split(' ')[1]) for s in self.neo_reader.header['spike_channels']]:

            all_channel_ids = self.neo_reader._get_channel_from_tetrode(tetrode_id)

            for i in range(4):
                tcmap[row_id, 0] = int(tetrode_id)
                tcmap[row_id, 1] = int(all_channel_ids[i])
                tcmap[row_id, 2] = int(i)
                row_id += 1

        del_idx = [False if i in channel_ids else True for i in tcmap[:, 1]]

        return np.delete(tcmap, del_idx, axis=0)



import setuptools

d = {}
exec(open("spikeextractors/version.py").read(), None, d)
version = d['version']
pkg_name = "spikeextractors"
long_description = open("README.md").read()

with open("full_requirements.txt", mode='r') as f:
    full_requires = f.read().split('\n')
full_requires = [e for e in full_requires if len(e) > 0]

extras_require = {"full": full_requires}

setuptools.setup(
    name=pkg_name,
    version=version,
    author="Alessio Buccino, Cole Hurwitz, Samuel Garcia, Jeremy Magland, Matthias Hennig",
    author_email="alessio.buccino@gmail.com",
    description="Python module for extracting recorded and spike sorted extracellular data from different file types and formats",
    url="https://github.com/SpikeInterface/spikeextractors",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    package_data={},
    include_package_data=True,
    install_requires=[
        'numpy',
        'tqdm',
        'joblib'
    ],
    extras_require=extras_require,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    )
)



from .neobaseextractor import NeoBaseRecordingExtractor

try:
    import neo

    HAVE_NEO = True
except ImportError:
    HAVE_NEO = False


class AxonaRecordingExtractor(NeoBaseRecordingExtractor):
    extractor_name = 'AxonaRecording'
    mode = 'file'
    NeoRawIOClass = 'AxonaRawIO'

    def __init__(self, **kargs):
        super().__init__(**kargs)

        # Read channel groups by tetrode IDs
        self.set_channel_groups(groups=[x - 1 for x in self.neo_reader.raw_annotations[
            'blocks'][0]['segments'][0]['signals'][0]['__array_annotations__']['tetrode_id']])

        header_channels = self.neo_reader.header['signal_channels'][slice(None)]

        names = header_channels['name']
        for i, ind in enumerate(self.get_channel_ids()):
            self.set_channel_property(channel_id=ind, property_name='name', value=names[i])



import shutil
import numpy as np
from pathlib import Path
from typing import Union, Optional

from spikeextractors import RecordingExtractor
from spikeextractors.extraction_tools import read_binary, write_to_binary_dat_format, check_get_traces_args

PathType = Union[str, Path]
DtypeType = Union[str, np.dtype]
ArrayType = Union[list, np.ndarray]
OptionalDtypeType = Optional[DtypeType]
OptionalArrayType = Optional[Union[np.ndarray, list]]


class BinDatRecordingExtractor(RecordingExtractor):
    """
    RecordingExtractor for a binary format

    Parameters
    ----------
    file_path: str or Path
        Path to the binary file
    sampling_frequency: float
        The sampling frequncy
    numchan: int
        Number of channels
    dtype: str or dtype
        The dtype of the binary file
    time_axis: int
        The axis of the time dimension (default 0: F order)
    recording_channels: list (optional)
        A list of channel ids
    geom: array-like (optional)
        A list or array with channel locations
    file_offset: int (optional)
        Number of bytes in the file to offset by during memmap instantiation.
    gain: float or array-like (optional)
        The gain to apply to the traces
    channel_offset: float or array-like
        The offset to apply to the traces
    is_filtered: bool
        If True, the recording is assumed to be filtered
    """
    extractor_name = 'BinDatRecording'
    has_default_locations = False
    has_unscaled = False
    installed = True
    is_writable = True
    mode = "file"
    installation_mesg = ""

    def __init__(self, file_path: PathType, sampling_frequency: float, numchan: int, dtype: DtypeType,
                 time_axis: int = 0, recording_channels: Optional[list] = None,  geom: Optional[ArrayType] = None,
                 file_offset: Optional[float] = 0,
                 gain: Optional[Union[float, ArrayType]] = None,
                 channel_offset: Optional[Union[float, ArrayType]] = None,
                 is_filtered: Optional[bool] = None):
        RecordingExtractor.__init__(self)
        self._datfile = Path(file_path)
        self._time_axis = time_axis
        self._dtype = np.dtype(dtype).name
        self._sampling_frequency = float(sampling_frequency)
        self._numchan = numchan
        self._geom = geom
        self._timeseries = read_binary(self._datfile, numchan, dtype, time_axis, file_offset)

        if is_filtered is not None:
            self.is_filtered = is_filtered
        else:
            self.is_filtered = False

        if recording_channels is not None:
            assert len(recording_channels) <= self._timeseries.shape[0], \
               'Provided recording channels have the wrong length'
            self._channels = recording_channels
        else:
            self._channels = list(range(self._timeseries.shape[0]))

        if len(self._channels) == self._timeseries.shape[0]:
            self._complete_channels = True
        else:
            assert max(self._channels) < self._timeseries.shape[0], "Channel ids exceed the number of " \
                                                                    "available channels"
            self._complete_channels = False

        if geom is not None:
            self.set_channel_locations(self._geom)
            self.has_default_locations = True

        if 'numpy' in str(dtype):
            dtype_str = str(dtype).replace("<class '", "").replace("'>", "")
            dtype_str = dtype_str.split('.')[1]
        else:
            dtype_str = str(dtype)

        if gain is not None:
            self.set_channel_gains(channel_ids=self.get_channel_ids(), gains=gain)
            self.has_unscaled = True

        if channel_offset is not None:
            self.set_channel_offsets(channel_offset)

        self._kwargs = {'file_path': str(Path(file_path).absolute()), 'sampling_frequency': sampling_frequency,
                        'numchan': numchan, 'dtype': dtype_str, 'recording_channels': recording_channels,
                        'time_axis': time_axis, 'geom': geom, 'file_offset': file_offset, 'gain': gain,
                        'is_filtered': is_filtered}

    def get_channel_ids(self):
        return self._channels

    def get_num_frames(self):
        return self._timeseries.shape[1]

    def get_sampling_frequency(self):
        return self._sampling_frequency

    @check_get_traces_args
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None, return_scaled=True):
        if self._complete_channels:
            if np.all(channel_ids == self.get_channel_ids()):
                traces = self._timeseries[:, start_frame:end_frame]
            else:
                channel_idxs = np.array([self.get_channel_ids().index(ch) for ch in channel_ids])
                if np.all(np.diff(channel_idxs) == 1):
                    traces = self._timeseries[channel_idxs[0]:channel_idxs[0]+len(channel_idxs),
                                              start_frame:end_frame]
                else:
                    # This block of the execution will return the data as an array, not a memmap
                    traces = self._timeseries[channel_idxs, start_frame:end_frame]
        else:
            # in this case channel ids are actually indexes
            traces = self._timeseries[channel_ids, start_frame:end_frame]
        return traces

    @staticmethod
    def write_recording(
        recording: RecordingExtractor,
        save_path: PathType,
        time_axis: int = 0,
        dtype: OptionalDtypeType = None,
        **write_binary_kwargs
    ):
        """
        Save the traces of a recording extractor in binary .dat format.

        Parameters
        ----------
        recording : RecordingExtractor
            The recording extractor object to be saved in .dat format.
        save_path : str
            The path to the file.
        time_axis : int, optional
            If 0 then traces are transposed to ensure (nb_sample, nb_channel) in the file.
            If 1, the traces shape (nb_channel, nb_sample) is kept in the file.
        dtype : dtype
            Type of the saved data. Default float32.
        **write_binary_kwargs: keyword arguments for write_to_binary_dat_format() function
        """
        write_to_binary_dat_format(recording, save_path, time_axis=time_axis, dtype=dtype,
                                   **write_binary_kwargs)



from pathlib import Path

from .cacheextractors import CacheRecordingExtractor, CacheSortingExtractor
from .recordingextractor import RecordingExtractor
from .sortingextractor import SortingExtractor


def save_si_object(object_name: str, si_object, output_folder,
                   cache_raw=False, include_properties=True, include_features=False):
    """
    Save an arbitrary SI object to a temprary location.

    Parameters
    ----------
    object_name: str
        The unique name of the SpikeInterface object.
    si_object: RecordingExtractor or SortingExtractor
        The extractor to be saved.
    output_folder: str or Path
        The folder where the object is saved.
    cache_raw: bool
        If True, the Extractor is cached to a binary file (not recommended for RecordingExtractor objects)
        (default False).
    include_properties: bool
        If True, properties (channel or unit) are saved (default True).
    include_features: bool
        If True, spike features are saved (default False)
    """
    Path(output_folder).mkdir(parents=True, exist_ok=True)

    if isinstance(si_object, RecordingExtractor):
        if not si_object.is_dumpable:
            cache = CacheRecordingExtractor(si_object, save_path=output_folder / "raw.dat")
        elif cache_raw:
            # save to json before caching to keep history (in case it's needed)
            json_file = output_folder / f"{object_name}.json"
            si_object.dump_to_json(output_folder / json_file)
            cache = CacheRecordingExtractor(si_object, save_path=output_folder / "raw.dat")
        else:
            cache = si_object

    elif isinstance(si_object, SortingExtractor):
        if not si_object.is_dumpable:
            cache = CacheSortingExtractor(si_object, save_path=output_folder / "sorting.npz")
        elif cache_raw:
            # save to json before caching to keep history (in case it's needed)
            json_file = output_folder / f"{object_name}.json"
            si_object.dump_to_json(output_folder / json_file)
            cache = CacheSortingExtractor(si_object, save_path=output_folder / "sorting.npz")
        else:
            cache = si_object
    else:
        raise ValueError("The 'si_object' argument shoulde be a SpikeInterface Extractor!")

    pkl_file = output_folder / f"{object_name}.pkl"
    cache.dump_to_pickle(
        output_folder / pkl_file,
        include_properties=include_properties,
        include_features=include_features
    )



from spikeextractors.extractors.bindatrecordingextractor import BinDatRecordingExtractor
import numpy as np
from pathlib import Path
import warnings

try:
    import xmltodict
    HAVE_XMLTODICT = True
except ImportError:
    HAVE_XMLTODICT = False


class NeuropixelsDatRecordingExtractor(BinDatRecordingExtractor):
    """
    Read raw Neurpoixels recordings from Open Ephys dat file and settings.xml
    
    This extractor is currently compatible with the 960 channel Neuropixels probes,
    where a maximum of 384 channels are recorded simulatenously. The array 
    configuration can be specified by passing the settings.xml file created by 
    OpenEphys (it can be found in the directory tree with teh recordings). If this 
    is not provided, the default configuration using 384 channels at the probe tip 
    will be used (a is warning printed).
   
    Parameters
    ----------
    file_path: str
        The raw data file (usually continuous.dat)
    settings_file: None or str
        The file settings.xml generated by OpenEphys containing the array 
        configuration. If not provided the default configuration using 384 
        channels at the probe tip will be used.
    verbose: bool
        Print probe configuration
    
    """
    extractor_name = 'NeuropixelsDatRecording'
    has_default_locations = True
    has_unscaled = True
    installed = HAVE_XMLTODICT
    is_writable = False
    mode = 'file'
    installation_mesg = "To use the NeuropixelsDat extractor, install xmltodict: \n\n pip install xmltodict\n\n"

    def __init__(self, file_path, settings_file=None, is_filtered=None, verbose=False):
        assert self.installed, self.installation_mesg
        source_dir = Path(__file__).parent
        self._settings_file = settings_file
        datfile = Path(file_path)
        time_axis = 0
        dtype = 'int16'
        sampling_frequency = float(30000)

        channel_locations = np.loadtxt(source_dir / 'channel_positions_neuropixels.txt')
        if self._settings_file is not None:
            with open(self._settings_file) as f:
                xmldata = f.read()
                settings = xmltodict.parse(xmldata)['SETTINGS']
            channel_info = settings['SIGNALCHAIN']['PROCESSOR'][0]['CHANNEL_INFO']
            channels = settings['SIGNALCHAIN']['PROCESSOR'][0]['CHANNEL']
            recorded_channels = []
            for c in channels:
                if c['SELECTIONSTATE']['@record'] == '1':
                    recorded_channels.append(int(c['@number']))
            used_channels = []
            used_channel_gains = []
            for c in channel_info['CHANNEL']:
                if 'AP' in c['@name'] and int(c['@number']) in recorded_channels:
                    used_channels.append(int(c['@number']))
                    used_channel_gains.append(float(c['@gain']))
            if verbose:
                print(f'{len(recorded_channels)} total channels found, with {len(used_channels)} recording AP')
                print(f'Channels used:\n{used_channels}')
            numchan = len(used_channels)
            geom = channel_locations[:, np.array(used_channels)].T
            gain = used_channel_gains[0]
            channels = used_channels
        else:
            warnings.warn("No information about this recording available,"
                          "using a default of 384 channels at the probe tip."
                          "If the recording differs, use settings_file=settings.xml")
            numchan = 384
            geom = channel_locations[:, :384].T
            gain = None
            channels = range(384)

        BinDatRecordingExtractor.__init__(self, file_path=datfile, numchan=numchan, dtype=dtype,
                                          sampling_frequency=sampling_frequency, gain=gain, geom=geom,
                                          recording_channels=channels, time_axis=time_axis, is_filtered=is_filtered)

        self._kwargs = {'filename': str(Path(file_path).absolute()), 'settings_file': settings_file,
                        'is_filtered': is_filtered}
